#include "Database.h"

#include <Base\Containers\index_range.hpp>
#include <Base\Containers\range_for_std_containers.hpp>
#include <Base\IO\FileUtil.hpp>
#include <Base\IO\ReadStreamBinLex.h>
#include <Base\IO\ReadStreamBinLex_stdvector.h>
#include <Base\Math\math_strfmt.h>
#include <Base\String\Formatting.h>
#include <Base\String\Trimming.h>

namespace hyplex::drakan
{
	static const bool kEnableDump = true;

	static void print(const FormattedTextOutputContext& ctx, const Model::Polygon& x)
	{
		printfmt(*ctx.stream, "%i %04x %02x %04x % .7f", x.n, x.unk, x.texId, asArrayView(x.indices), asArrayView(x.uvs));
	}

	namespace
	{
		struct ColladaMat4
		{
			fmat4 x;
		};

		static void print(const FormattedTextOutputContext& ctx, const ColladaMat4& x)
		{
			for (unsigned int r = 0; r < 4; ++r)
			{
				for (unsigned int c = 0; c < 4; ++c)
				{
					print(ctx, x.x.axes[c][r]);
					if(c < 3)
						ctx.stream->write(' ');
				}
				if(r < 3)
					ctx.stream->write(' ');
			}
		}
	}

	namespace
	{
		template<class T>
		struct IndentedArray
		{
			array_view<T> items;
		};

		template<class C>
		auto makeIndentedArray(C&& c)
		{
			auto av = asArrayView(c);
			return IndentedArray<decltype(av)::value_type>{ av };
		}

		template<class T>
		void print(const FormattedTextOutputContext& ctx, IndentedArray<T> array)
		{
			ctx.stream->write("[\n"_sv);
			for (size_t i = 0; i < array.items.size(); ++i)
			{
				ctx.stream->write('\t');
				print(ctx, i);
				ctx.stream->write(' ');
				print(ctx, array.items[i]);
				ctx.stream->write(",\n"_sv);
			}
			ctx.stream->write("]"_sv);
		}
	}

	void Model::consumeData(const Entry& entry, BasicBufferedLexState& bin)
	{
		if (entry.type == 0)
		{ // Model info
			name = trimNul(readString(bin, read<uint16_t>(bin)));
			flags = read<uint32_t>(bin);
		}
		else if (entry.type == 3)
			readArray(bin, vertices, read<uint16_t>(bin));
		else if (entry.type == 4)
		{
			const unsigned int numPolys = read<uint16_t>(bin);
			polys.clear();
			polys.reserve(numPolys);

			for (unsigned int i = 0; i < numPolys; ++i)
			{
				const struct PolyHeader
				{
					uint16_t unk;
					uint16_t n;
					uint16_t texId;
				} polyHeader = read<PolyHeader>(bin);

				HYPLEX_ENFORCE(polyHeader.n >= 3 && polyHeader.n <= 4);

				std::array<uint16_t, 4> indices{};
				std::array<fvec2, 4> uvs{};

				for (unsigned int j = 0; j < polyHeader.n; ++j)
				{
					read(bin, indices[j]);
					read(bin, uvs[j]);
				}

				Model::Polygon poly{};
				poly.n = polyHeader.n;
				poly.unk = polyHeader.unk;
				poly.texId = polyHeader.texId;
				poly.indices = indices;
				poly.uvs = uvs;
				polys.push_back(poly);
			}
		}
		else if (entry.type == 6)
			readArray(bin, textureIds, read<uint32_t>(bin));
		else if (entry.type == 8)
		{
			boundingInfo = read<BoundingInfo>(bin);

			const unsigned int numLods = read<uint16_t>(bin);
			lods.clear();
			lods.resize(numLods);
			if (numLods > 0)
				lods[0].name = "Default Unnamed LOD";
			for (unsigned int i = 1; i < numLods; ++i)
				lods[i].name = trimNul(readString(bin, read<uint16_t>(bin)));

			const unsigned int numNodes = read<uint16_t>(bin);
			nodes.clear();
			nodes.resize(numNodes);

			for (unsigned int i = 0; i < numNodes; ++i)
			{
				const struct Node
				{
					char name[32];
					int32_t parentI;
				} node = read<Node>(bin);
				if (node.parentI >= 0)
					nodes[node.parentI].name = cstringView(node.name);
			}
			const unsigned int numReducedNodes = read<uint16_t>(bin);

			HYPLEX_ENFORCE(numReducedNodes <= numNodes);

			nodes.resize(numReducedNodes);

			const unsigned int numEnvelopes = numLods;

			for (unsigned int i = 0; i < numReducedNodes; ++i)
			{
				const struct UnkStruct1
				{
					fmat3x4 xform;
					int32_t meshI;
					int32_t childI;
					int32_t siblingI;
				} unkNodeInfo = read<UnkStruct1>(bin);

				std::vector<VertexWeight> weights;
				for (unsigned int j = 0; j < numEnvelopes; ++j)
				{
					const unsigned int numWeights = read<uint16_t>(bin);
					if (j == 0)
						readArray(bin, weights, numWeights);
					else
						skipUptoN(bin, numWeights * sizeof(VertexWeight));
				}

				nodes[i].worldXForm = unkNodeInfo.xform.getOrthonormalInverse(); // Took a while to figure this out.
				nodes[i].meshI = unkNodeInfo.meshI;
				nodes[i].childI = unkNodeInfo.childI;
				nodes[i].siblingI = unkNodeInfo.siblingI;
				nodes[i].vertexWeights = std::move(weights);
			}

			for (unsigned int i = 0; i < numLods; ++i)
				readArray(bin, lods[i].meshes, read<uint16_t>(bin));

			readArray(bin, animIds, read<uint16_t>(bin));

			const unsigned int numChannels = read<uint16_t>(bin);
			channels.clear();
			channels.resize(numChannels);
			for (unsigned int i = 0; i < numChannels; ++i)
			{
				const unsigned int nodeI = read<int32_t>(bin);
				const fmat3x4 xformA = read<fmat3x4>(bin);
				const fmat3x4 xformB = read<fmat3x4>(bin);
				const unsigned int numLODCaps = read<uint16_t>(bin);

				for (unsigned int j = 0; j < numLODCaps; ++j)
				{
					Channel::LODCapInfo lodCapInfo;
					static_cast<Channel::RawLODCapInfo&>(lodCapInfo) = read<Channel::RawLODCapInfo>(bin);
					readArray(bin, lodCapInfo.weights, read<uint16_t>(bin));
					channels[i].lods.push_back(std::move(lodCapInfo));
				}

				channels[i].nodeI = nodeI;
				channels[i].xformA = xformA;
				channels[i].xformB = xformB;
			}
		}
	}

	static String escapeToHtml(string_view x)
	{
		String out;
		out.reserve(x.size());
		for (const char c : x)
		{
			if (isalnum(uint8_t(c)) || c == ':' || c == '/')
				out += c;
			else
				printfmt(out, "%%%02x", uint8_t(c));
		}
		return out;
	}

	String Model::buildCollada(const NodeDumpArg& arg) const
	{
		String str;
		str += R"(<?xml version="1.0" encoding="utf-8"?>)" "\n";
		str += R"(<COLLADA xmlns="http://www.collada.org/2005/11/COLLADASchema" version="1.4.1">)" "\n";
		str += R"(<asset>)" "\n";
		str += R"(	<unit name="meter" meter="1"/>)" "\n";
		str += R"(	<up_axis>Y_UP</up_axis>)" "\n";
		str += R"(</asset>)" "\n";

		str += R"(<library_lights>)" "\n";
		str += R"(	<light id="Point-light" name="Point">)" "\n";
		str += R"(		<technique_common>)" "\n";
		str += R"(			<directional>)" "\n";
		str += R"(				<color sid="color">5.41 5.41 5.41</color>)" "\n";
		str += R"(			</directional>)" "\n";
		str += R"(		</technique_common>)" "\n";
		str += R"(		<extra>)" "\n";
		str += R"(			<technique profile="blender">)" "\n";
		str += R"(				<energy sid="blender_energy">10</energy>)" "\n";
		str += R"(				<mode>4097</mode>)" "\n";
		str += R"(				<type>1</type>)" "\n";
		str += R"(			</technique>)" "\n";
		str += R"(		</extra>)" "\n";
			
		str += R"(	</light>)" "\n";
		str += R"(</library_lights>)" "\n";

		for (size_t texI = 0; texI < textureIds.size(); ++texI)
		{
			String textureDir;
			if (const unsigned int dbId = textureIds[texI] >> 16)
			{
				textureDir = arg.outputRootDir;
				bool found = false;
				for (const Dependency& dep : arg.dbMeta->dependencies)
				{
					if (dep.id == dbId)
					{
						found = true;
						textureDir += string_view(dep.path);
						break;
					}
				}
				HYPLEX_ENFORCE(found);
				textureDir += " [root]/"_sv;
			}
			else
				textureDir = arg.dbRootDir;

			str += R"(<library_images>)" "\n";
			printfmt(str, R"(	<image id="Image%u">)" "\n", texI);
			printfmt(str, R"(		<init_from>/%sTexture%u.png</init_from>)" "\n", escapeToHtml(textureDir), textureIds[texI] & 0xFFFF);
			str += R"(	</image>)" "\n";
			str += R"(</library_images>)" "\n";

			str += R"(<library_effects>)" "\n";
			printfmt(str, R"(	<effect id="Material%u-effect">)" "\n", texI);
			str += R"(		<profile_COMMON>)" "\n";
			printfmt(str, R"(			<newparam sid="Image%u-surface">)" "\n", texI);
			str += R"(				<surface type="2D">)" "\n";
			printfmt(str, R"(					<init_from>Image%u</init_from>)" "\n", texI);
			str += R"(				</surface>)" "\n";
			str += R"(			</newparam>)" "\n";
			printfmt(str, R"(			<newparam sid="Image%u-sampler">)" "\n", texI);
			str += R"(				<sampler2D>)" "\n";
			printfmt(str, R"(					<source>Image%u-surface</source>)" "\n", texI);
			str += R"(				</sampler2D>)" "\n";
			str += R"(			</newparam>)" "\n";
			str += R"(			<technique sid="common">)" "\n";
			str += R"(				<lambert>)" "\n";
			// Blender will export this but won't import it properly.
			str += R"(					<emission>)" "\n";
			str += R"(					<color sid="emission">0 0 0 1</color>)" "\n";
			str += R"(					</emission>)" "\n";
			str += R"(					<ambient>)" "\n";
			str += R"(					<color sid="ambient">0 0 0 1</color>)" "\n";
			str += R"(					</ambient>)" "\n";
			str += R"(					<transparent>)" "\n";
			printfmt(str, R"(						<texture texture="Image%u-sampler" texcoord="UVMap"/>)" "\n", texI);
			str += R"(					</transparent>)" "\n";
			str += R"(					<transparency>)" "\n";
			str += R"(						<float sid="transparency">0</float>)" "\n";
			str += R"(					</transparency>)" "\n";
			str += R"(					<diffuse>)" "\n";
			printfmt(str, R"(						<texture texture="Image%u-sampler" texcoord="UVMap"/>)" "\n", texI);
			str += R"(					</diffuse>)" "\n";
			str += R"(				</lambert>)" "\n";
			str += R"(			</technique>)" "\n";
			str += R"(		</profile_COMMON>)" "\n";
			str += R"(	</effect>)" "\n";
			str += R"(</library_effects>)" "\n";
			str += R"(<library_materials>)" "\n";
			printfmt(str, R"(		<material id="Material%u-material">)" "\n", texI);
			printfmt(str, R"(			<instance_effect url="#Material%u-effect"/>)" "\n", texI);
			str += R"(		</material>)" "\n";
			str += R"(</library_materials>)" "\n";
		}

		str += R"(<library_geometries>)" "\n";
		str += R"(	<geometry id="Cube_001-mesh" name="Cube.001">)" "\n";
		str += R"(		<mesh>)" "\n";
		str += R"(			<source id="Cube_001-mesh-positions">)" "\n";
		str += R"(				<float_array id="Cube_001-mesh-positions-array" count=")";
		printfmt(str, "%u\">", vertices.size() * 3);
		for (const fvec3& v : vertices)
			printfmt(str, "%g %g %g ", v.x, v.y, v.z);
		str += "</float_array>\n";
		str += R"(				<technique_common>)" "\n";
		printfmt(str, R"(					<accessor source="#Cube_001-mesh-positions-array" count="%u" stride="3">)" "\n", vertices.size());
		str += R"(						<param name="X" type="float"/>)" "\n";
		str += R"(						<param name="Y" type="float"/>)" "\n";
		str += R"(						<param name="Z" type="float"/>)" "\n";
		str += R"(					</accessor>)" "\n";
		str += R"(				</technique_common>)" "\n";
		str += R"(			</source>)" "\n";
		str += R"(			<source id="Cube-mesh-map-0">)" "\n";

		array_view<const Polygon> lodPolys = polys;
		if (lods.size() > 0 && lods[0].meshes.size() > 0)
			lodPolys = lodPolys.slice_n(lods[0].meshes[0].firstPolyI, lods[0].meshes[0].numPolys);

		size_t totalUVs = 0;
		for (const Polygon& poly : lodPolys)
			totalUVs += poly.n;
		printfmt(str, R"(				<float_array id="Cube-mesh-map-0-array" count="%u">)", totalUVs * 2);
		for (const Polygon& poly : lodPolys)
		{
			for (unsigned int i = 0; i < poly.n; ++i)
				printfmt(str, "%g %g  ", poly.uvs[i].x, poly.uvs[i].y);
		}
		str += R"(</float_array>)" "\n";
		str += R"(				<technique_common>)" "\n";
		printfmt(str, R"(					<accessor source="#Cube-mesh-map-0-array" count="%u" stride="2">)" "\n", totalUVs);
		str += R"(						<param name="S" type="float"/>)" "\n";
		str += R"(						<param name="T" type="float"/>)" "\n";
		str += R"(					</accessor>)" "\n";
		str += R"(				</technique_common>)" "\n";
		str += R"(			</source>)" "\n";
		str += R"(			<vertices id="Cube_001-mesh-vertices">)" "\n";
		str += R"(				<input semantic="POSITION" source="#Cube_001-mesh-positions"/>)" "\n";
		str += R"(			</vertices>)" "\n";

		std::vector<index_range> invariantMaterialPolyIRs;
		for (unsigned int i = 0; i < lodPolys.size(); ++i)
			if (i && lodPolys[i - 1].texId != lodPolys[i].texId)
				invariantMaterialPolyIRs.push_back(index_range(invariantMaterialPolyIRs.empty() ? 0 : invariantMaterialPolyIRs.back().end(), i));
		if(!lodPolys.empty())
			invariantMaterialPolyIRs.push_back(index_range(invariantMaterialPolyIRs.empty() ? 0 : invariantMaterialPolyIRs.back().end(), lodPolys.size()));

		size_t uvI = 0;

		for (const index_range polyIR : invariantMaterialPolyIRs)
		{
			const auto submeshPolys = polyIR.in(lodPolys);

			printfmt(str, R"(			<polylist material="SubMeshMaterial%u" count="%u">)" "\n", polyIR.begin(), submeshPolys.size());
			str += R"(				<input semantic="VERTEX" source="#Cube_001-mesh-vertices" offset="0"/>)" "\n";
			str += R"(				<input semantic="TEXCOORD" source="#Cube-mesh-map-0" offset="1" set="0"/>)" "\n";
			str += R"(				<vcount>)";

			for (const Polygon& poly : submeshPolys)
				printfmt(str, "%u ", poly.n);
			str += "</vcount>\n";
			str += R"(				<p>)";

			
			for (const Polygon& poly : submeshPolys)
			{
				for (unsigned int i = 0; i < poly.n; ++i)
				{
					printfmt(str, "%u ", poly.indices[i]);
					printfmt(str, "%u ", uvI++);
				}
			}
			str += "</p>\n";
			str += R"(			</polylist>)" "\n";
		}
		str += R"(		</mesh>)" "\n";
		str += R"(	</geometry>)" "\n";
		str += R"(</library_geometries>)" "\n";


		str += R"(<library_controllers>)" "\n";
		str += R"(	<controller id="Armature_Cube-skin" name="Armature">)" "\n";
		str += R"(		<skin source="#Cube_001-mesh">)" "\n";
		str += R"(			<bind_shape_matrix>1 0 0 0 0 1 0 0 0 0 1 0 0 0 0 1</bind_shape_matrix>)" "\n";
		str += R"(			<source id="Armature_Cube-skin-joints">)" "\n";
		printfmt(str, R"(				<Name_array id="Armature_Cube-skin-joints-array" count="%u">)" "\n", nodes.size());
		for (unsigned int i = 0; i < nodes.size(); ++i)
			printfmt(str, "Bone%u ", i);
		str += R"(</Name_array>)" "\n";
		str += R"(				<technique_common>)" "\n";
		printfmt(str, R"(					<accessor source="#Armature_Cube-skin-joints-array" count="%u" stride="1">)" "\n", nodes.size());
		str += R"(						<param name="JOINT" type="name"/>)" "\n";
		str += R"(					</accessor>)" "\n";
		str += R"(				</technique_common>)" "\n";
		str += R"(			</source>)" "\n";
		str += R"(			<source id="Armature_Cube-skin-bind_poses">)" "\n";
		printfmt(str, R"(				<float_array id="Armature_Cube-skin-bind_poses-array" count="%u">)" "\n", nodes.size() * 16);
		for (unsigned int i = 0; i < nodes.size(); ++i)
			printfmt(str, "					%g\n", ColladaMat4{ nodes[i].worldXForm.getOrthonormalInverse() });
		str += R"(				</float_array>)" "\n";
		str += R"(				<technique_common>)" "\n";
		printfmt(str, R"(					<accessor source="#Armature_Cube-skin-bind_poses-array" count="%u" stride="16">)" "\n", nodes.size());
		str += R"(						<param name="TRANSFORM" type="float4x4"/>)" "\n";
		str += R"(					</accessor>)" "\n";
		str += R"(				</technique_common>)" "\n";
		str += R"(			</source>)" "\n";
		str += R"(			<source id="Armature_Cube-skin-weights">)" "\n";

		struct VertexJointWeight
		{
			unsigned int vertexI;
			unsigned int nodeI;
			float weight;

			bool operator<(const VertexJointWeight& b) const noexcept
			{
				return vertexI < b.vertexI;
			}
		};
		std::vector<VertexJointWeight> vertexJointWeights;
		for (unsigned int i = 0; i < nodes.size(); ++i)
			for (const VertexWeight& vw : nodes[i].vertexWeights)
				vertexJointWeights.push_back(VertexJointWeight{ vw.vertexI, i, vw.weight });
		std::sort(vertexJointWeights.begin(), vertexJointWeights.end());

		printfmt(str, R"(				<float_array id="Armature_Cube-skin-weights-array" count="%u">)", vertexJointWeights.size());
		unsigned int numVerticesOnSkin = 0;
		for (unsigned int i = 0; i < vertexJointWeights.size(); ++i)
		{
			const unsigned int vertexI = vertexJointWeights[i].vertexI;
			numVerticesOnSkin = std::max(numVerticesOnSkin, vertexI + 1);
			if (i <= 0 || vertexI != vertexJointWeights[i - 1].vertexI)
				printfmt(str, "\n					<!-- % 4u --> ", vertexI);
			printfmt(str, "%g ", vertexJointWeights[i].weight);
		}
		numVerticesOnSkin = (unsigned int)vertices.size();
		str += "\n" R"(				</float_array>)" "\n";
		str += R"(				<technique_common>)" "\n";
		printfmt(str, R"(					<accessor source="#Armature_Cube-skin-weights-array" count="%u" stride="1">)" "\n", vertexJointWeights.size());
		str += R"(						<param name="WEIGHT" type="float"/>)" "\n";
		str += R"(					</accessor>)" "\n";
		str += R"(				</technique_common>)" "\n";
		str += R"(			</source>)" "\n";
		str += R"(			<joints>)" "\n";
		str += R"(				<input semantic="JOINT" source="#Armature_Cube-skin-joints"/>)" "\n";
		str += R"(				<input semantic="INV_BIND_MATRIX" source="#Armature_Cube-skin-bind_poses"/>)" "\n";
		str += R"(			</joints>)" "\n";
		printfmt(str, R"(			<vertex_weights count="%u">)" "\n", numVerticesOnSkin);
		str += R"(				<input semantic="JOINT" source="#Armature_Cube-skin-joints" offset="0"/>)" "\n";
		str += R"(				<input semantic="WEIGHT" source="#Armature_Cube-skin-weights" offset="1"/>)" "\n";
		str += R"(				<vcount>)";
		for (unsigned int vi = 0, i = 0; vi < numVerticesOnSkin; ++vi)
		{
			unsigned int j = i;
			while (j < vertexJointWeights.size() && vertexJointWeights[j].vertexI <= vi)
				++j;
			printfmt(str, "%u ", j - i);
			i = j;
		}
		str += R"(</vcount>)" "\n";
		str += R"(				<v>)";
		for (unsigned int i = 0; i < vertexJointWeights.size(); ++i)
		{
			const unsigned int vertexI = vertexJointWeights[i].vertexI;
			numVerticesOnSkin = std::max(numVerticesOnSkin, vertexI + 1);
			if (i <= 0 || vertexI != vertexJointWeights[i - 1].vertexI)
				printfmt(str, "\n					<!-- % 4u --> ", vertexI);
			printfmt(str, "%u %u  ", vertexJointWeights[i].nodeI, i);
		}
		str += "\n" R"(				</v>)" "\n";
		str += R"(			</vertex_weights>)" "\n";
		str += R"(		</skin>)" "\n";
		str += R"(	</controller>)" "\n";
		str += R"(</library_controllers>)" "\n";

		str += R"(<library_visual_scenes>)" "\n";
		str += R"(	<visual_scene id="Scene" name="Scene">)" "\n";
		str += R"(		<node id="Armature" name="Armature" type="NODE">)" "\n";
		str += R"(			<matrix sid="transform">1 0 0 0 0 1 0 0 0 0 1 0 0 0 0 1</matrix>)" "\n";

		struct NodeHierarchyOutput
		{
			array_view<const SceneNode> nodes;

			void operator()(String& out, unsigned int nodeI, const fmat3x4& parentXForm, unsigned int level = 0) const
			{
				(void)parentXForm;
				out.append(level + 1, '\t');
				// sid needed for skin controller.
				printfmt(out, R"(<node id="Bone%u" sid="Bone%u" name="%s" type="JOINT">)" "\n", nodeI, nodeI, nodes[nodeI].name);

				const fmat3x4 globalXForm = nodes[nodeI].worldXForm;
				const fmat3x4 localXForm = globalXForm * parentXForm.getOrthonormalInverse();

				const fmat3 globalRot = globalXForm.linear;
				const fvec3 globalTranslation = globalXForm.translation;
				//const fvec3 localTranslation = parentXForm.getOrthonormalInverse() * globalTranslation;
				//const fmat3 localRot = parentXForm.get3x3Part().getOrthonormalInverse() * globalRot;
				const fvec3 localTranslation = localXForm.translation;
				const fmat3 localRot = localXForm.linear;
				
				out.append(level + 1, '\t');
				/*printfmt(out, R"(	<matrix sid="transform">%g %g %g 0 %g %g %g 0 %g %g %g 0 %g %g %g 1</matrix>)" "\n",
					xform.getXAxis()[0], xform.getXAxis()[1], xform.getXAxis()[2],
					xform.getYAxis()[0], xform.getYAxis()[1], xform.getYAxis()[2],
					xform.getZAxis()[0], xform.getZAxis()[1], xform.getZAxis()[2],
					xform.getTranslation()[0], xform.getTranslation()[1], xform.getTranslation()[2]);
					*/

				//const fvec3 translation = parentXForm * nodes[nodeI].xform.getTranslation();
				//xform.translation = translation;
				//const fvec3 globalTranslation = localXForm.linear * nodes[nodeI].xform.translation;
				//const fvec3 localTranslation = parentXForm.getOrthonormalInverse() * globalTranslation;

				printfmt(out, R"(	<matrix sid="transform">%g</matrix>)" "\n", ColladaMat4{ fmat3x4(localRot, localTranslation) });


				unsigned int childI = nodes[nodeI].childI;
				while (childI > 0)
				{
					//(*this)(out, childI, fmat3x4(nodes[nodeI].xform.linear, translation), level + 1);
					(*this)(out, childI, fmat3x4(globalRot, globalTranslation), level + 1);
					childI = nodes[childI].siblingI;
				}

				out.append(level + 1, '\t');
				printfmt(out, R"(</node>)" "\n");
			}
		};

		if(!nodes.empty())
			NodeHierarchyOutput{ nodes }(str, 0, identity);
		

		//str += R"(			<node id="Bone" name="Bone" sid="Bone" type="JOINT">)" "\n";
		//str += R"(				<matrix sid="transform">1 0 0 0 0 0 -1 0 0 1 0 -0.7515755 0 0 0 1</matrix>)" "\n";
		//str += R"(				<node id="Bone_001" name="Bone.001" sid="Bone_001" type="JOINT">)" "\n";
		//str += R"(					<matrix sid="transform">0.7006941 0.1752424 -0.6916052 0 -0.2773153 0.960039 -0.03770015 1 0.6573614 0.218209 0.7212912 0 0 0 0 1</matrix>)" "\n";
		//str += R"(				</node>)" "\n";
		//str += R"(			</node>)" "\n";
		str += R"(		</node>)" "\n";
		str += R"(		<node id="Cube" name="Cube" type="NODE">)" "\n";
		str += R"(			<translate sid="location">0 0 0</translate>)" "\n";
		str += R"(			<rotate sid="rotationX">1 0 0 0</rotate>)" "\n";
		str += R"(			<rotate sid="rotationY">0 1 0 0</rotate>)" "\n";
		str += R"(			<rotate sid="rotationZ">0 0 1 0</rotate>)" "\n";
		str += R"(			<scale sid="scale">1 1 1</scale>)" "\n";
		str += R"(			<instance_controller url="#Armature_Cube-skin">)" "\n";
		str += R"(				<skeleton>#Bone0</skeleton>)" "\n";
		str += R"(				<bind_material>)" "\n";
		str += R"(					<technique_common>)" "\n";
		for (const index_range polyIR : invariantMaterialPolyIRs)
		{
			printfmt(str, R"(						<instance_material symbol="SubMeshMaterial%u" target="#Material%u-material">)" "\n", polyIR.begin(), lodPolys[polyIR.begin()].texId);
			str += R"(							<bind_vertex_input semantic="UVMap" input_semantic="TEXCOORD" input_set="0"/>)" "\n";
			str += R"(						</instance_material>)" "\n";
		}
		str += R"(					</technique_common>)" "\n";
		str += R"(				</bind_material>)" "\n";
		str += R"(			</instance_controller>)" "\n";
		str += R"(		</node>)" "\n";
		str += R"(		<node id="Point" name="Point" type="NODE">)" "\n";
		str += R"(			<translate sid="location">0 1 -1</translate>)" "\n";
		str += R"(			<rotate sid="rotationX">1 0 0 -135</rotate>)" "\n";
		str += R"(			<rotate sid="rotationY">0 1 0 45</rotate>)" "\n";
		str += R"(			<instance_light url="#Point-light"/>)" "\n";
		str += R"(		</node>)" "\n";
		str += R"(	</visual_scene>)" "\n";
		str += R"(</library_visual_scenes>)" "\n";
		str += R"(<scene>)" "\n";
		str += R"(	<instance_visual_scene url="#Scene"/>)" "\n";
		str += R"(</scene>)" "\n";

		str += "</COLLADA>";

		return str;
	}

	ENodeType Model::getType() const noexcept
	{
		return ENodeType::MODEL;
	}

	static void print(const FormattedTextOutputContext& ctx, const Model::VertexWeight& x)
	{
		printfmt(*ctx.stream, "%u=%.2f", x.vertexI, x.weight);
	}

	static void print(const FormattedTextOutputContext& ctx, const Model::SceneNode& x)
	{
		//printfmt(*ctx.stream, "name=\"%s\" meshI=%i childI=%i siblingI=%i xform=%g vertexWeights=%s",
		//	x.name, x.meshI, x.childI, x.siblingI, x.xform, asArrayView(x.vertexWeights));
		printfmt(*ctx.stream, "name=\"%s\" meshI=%i childI=%i siblingI=%i xform=%g",
			x.name, x.meshI, x.childI, x.siblingI, x.worldXForm);
	}


	void Model::dumpNode(const NodeDumpArg& arg) const
	{
		if (!kEnableDump)
			return;

		String str;

		printfmt(str, "name %s\n", name);
		printfmt(str, "id %04x\n", id);
		printfmt(str, "flags %08x\n", flags);
		printfmt(str, "vertices %g\n", makeIndentedArray(vertices));
		printfmt(str, "textures %04x\n", asArrayView(textureIds));
		printfmt(str, "polys %s\n", makeIndentedArray(polys));

		printfmt(str, "boundingInfo %g %g %g %g\n",
			fvec4(boundingInfo.boundingSphereCenter, boundingInfo.boundingSphereRadius) * 2048.0f,
			fvec4(boundingInfo.unk0[0], boundingInfo.unk0[1], boundingInfo.unk0[2], boundingInfo.boundingBoxSizeX) * 2048.0f,
			fvec4(boundingInfo.unk1[0], boundingInfo.unk1[1], boundingInfo.unk1[2], boundingInfo.boundingBoxSizeY) * 2048.0f,
			fvec4(boundingInfo.unk2[0], boundingInfo.unk2[1], boundingInfo.unk2[2], boundingInfo.boundingBoxSizeZ) * 2048.0f);
		printfmt(str, "nodes %s\n", makeIndentedArray(nodes));

		for (size_t i = 0; i < lods.size(); ++i)
		{
			printfmt(str, "lod %u %s\n", i, lods[i].name);
			for (const LODInfo::Mesh& mesh : lods[i].meshes)
			{
				printfmt(str, "\tMesh meshI=%i unk=%i usage=%i nodeI=%i firstVertexI=%i numVertices=%i firstPolyI=%i numPolys=%i\n",
					mesh.meshI, mesh.unk, mesh.usage, mesh.nodeI, mesh.firstVertexI, mesh.numVertices, mesh.firstPolyI, mesh.numPolys);
			}
		}
		
		printfmt(str, "anims %04x\n", asArrayView(animIds));

		for (size_t i = 0; i < channels.size(); ++i)
		{
			const Channel& channel = channels[i];
			printfmt(str, "channel %u nodeI=%i xformA=%g xformB=%g\n", i, channel.nodeI, channel.xformA, channel.xformB);
			for (const Channel::LODCapInfo& cap : channel.lods)
			{
				printfmt(str, "\tLODCapInfo firstCapPolyI=%i numCapPolys=%i firstPartPolyI=%i numPartPolys=%i unk=%i vertexWeights=%.2g\n",
					cap.firstCapPolyI, cap.numCapPolys, cap.firstPartPolyI, cap.numPartPolys, cap.unk, asArrayView(cap.weights));
			}
		}

		const String pathWithoutExt = sprintfmt("%sModel%u %s", arg.dir, id, escapeToFilename(name));

		FileUnbufferedWriteStream(pathWithoutExt + ".txt"_sv, EFileWriteOpenMode::RECREATE, true)
			.write(asRawUChars(asArrayView(str)));

		FileUnbufferedWriteStream(pathWithoutExt + ".dae"_sv, EFileWriteOpenMode::RECREATE, true)
			.write(asRawUChars(asArrayView(buildCollada(arg))));
	}
}