#include "Database.h"
#include "RFL.h"
#include "FilesystemPathUtil.h"

#include <Base\Containers\range_for_std_containers.hpp>
#include <Base\IO\FileUtil.hpp>
#include <Base\IO\ReadStreamBinLex.h>
#include <Base\IO\ReadStreamBinLex_stdvector.h>
#include <Base\Math\vec3.h>
#include <Base\Math\mat3x4.h>
#include <Base\Math\math_strfmt.h>
#include <Base\String\Formatting.h>
#include <Base\String\Trimming.h>

#include <vector>
#include <iostream>
#include <fstream>
#include <string_view>
#include <map>
#include <filesystem>
#include <codecvt>

using namespace hyplex;
namespace fs = std::experimental::filesystem;

#pragma pack(push)
#pragma pack(1)
struct Header
{
	char id[4];
	uint16_t version;
	uint32_t directoryOffset;
	uint16_t numEntries;
};

struct Entry
{
	uint8_t type;
	uint8_t subtype;
	uint16_t unk00;
	uint16_t parentDirectory;
	uint32_t offset;
	uint32_t size;
};
#pragma pack(pop)

static const string_view kInputRFLPath = R"(.\Dragon.rfl)";
static const string_view kRootInputDir = R"(.\)";
static const string_view kRootOutputDir = R"(.\Drakan Dump\)";

namespace
{
	struct CaseInsensitiveStringComparator
	{
		bool operator()(std::string_view a, std::string_view b) const noexcept
		{
			if (a.size() != b.size())
				return a.size() < b.size();
			else
				return _strnicmp(a.data(), b.data(), a.size()) < 0;
		}

		bool operator()(std::wstring_view a, std::wstring_view b) const noexcept
		{
			if (a.size() != b.size())
				return a.size() < b.size();
			else
				return _wcsnicmp(a.data(), b.data(), a.size()) < 0;
		}
	};

	std::optional<unsigned int> consumeUInt(BasicBufferedLexState& lex) noexcept
	{
		unsigned int x = 0;
		int n = 0;
		if (_snscanf_s(lex.buf.data(), lex.buf.size(), "%u%n", &x, &n) == 1)
		{
			lex.buf.pop_front(n);
			return x;
		}
		else
			return std::nullopt;
	}

	bool skipWhitespace(BasicBufferedLexState& lex) noexcept
	{
		return skipWhile(lex, isAsciiSpace) > 0;
	}
}

int main() try
{
	

	{
		fs::create_directories(fs::path(kRootOutputDir.begin(), kRootOutputDir.end()));
		drakan::extractRFLResources(kInputRFLPath, kRootOutputDir);

		const auto classCats = drakan::loadRFLClassCategories(asRawUChars(asArrayView(loadEntireFileAsString(kRootOutputDir + "RFL Types.bin"_sv))));
		const auto classes = drakan::loadRFLClasses(asRawUChars(asArrayView(loadEntireFileAsString(kRootOutputDir + "RFL Objects.bin"_sv))));

		String rflDump;
		for (const auto& kv : classCats)
			printfmt(rflDump, "%u %s\n", kv.first, kv.second);
		rflDump += '\n';
		for (const auto& kv : classes)
			rflDump += kv.second.toString(classCats, classes);

		FileUnbufferedWriteStream(kRootOutputDir + "RFL Dump.txt"_sv, EFileWriteOpenMode::RECREATE, true)
			.write(asRawUChars(asArrayView(rflDump)));
	}

	//fs::create_directory(fs::path(kRootOutputDir.begin(), kRootOutputDir.end()));
	//dumpDatabase(R"(Common\NPC\NPCs.mod)");
	//dumpDatabase(R"(music.rrc)");

	std::map<std::string, drakan::DatabaseCollection, CaseInsensitiveStringComparator> dbLookup;

	const fs::path rootInputPath = fs::path(kRootInputDir.begin(), kRootInputDir.end());

	for (const fs::path& path : fs::recursive_directory_iterator(rootInputPath))
	{
		if (path.extension() != L".db")
			continue;

		//const fs::path relDBDir = lexically_relative(path.parent_path(), rootInputPath);

		const String file = loadEntireFileAsString(string_view(path.generic_string()));

		BasicBufferedLexState lex{ file };

		unsigned int version = 0;
		int n = 0;
		if (_snscanf_s(lex.buf.data(), lex.buf.size(), "version %u\n%n", &version, &n) != 1)
		{
			//throw std::runtime_error("Expected version.");
			continue;
		}
		lex.buf.pop_front(n);
		unsigned int numDependencies = 0;
		if(_snscanf_s(lex.buf.data(), lex.buf.size(), "dependencies %u\n%n", &numDependencies, &n) != 1)
			throw std::runtime_error("Expected dependencies.");
		lex.buf.pop_front(n);

		drakan::DatabaseCollection dbColl;

		for (unsigned int i = 0; i < numDependencies; ++i)
		{
			skipWhitespace(lex);

			if (!peek(lex))
				break;

			const std::optional<unsigned int> id = consumeUInt(lex);
			const bool hasWS = skipWhitespace(lex);
			String dependencyPath;
			if(!id || !hasWS)
				throw std::runtime_error("Expected dependency id and path.");

			consumeLineAndCRLF(lex, dependencyPath);

			if(dependencyPath.empty())
				throw std::runtime_error("Expected dependency path.");

			std::wstring_convert<std::codecvt_utf8_utf16<wchar_t>> converter;
			std::wstring wide = converter.from_bytes(dependencyPath.data(), dependencyPath.data() + dependencyPath.size());

			dbColl.dependencies.push_back(drakan::Dependency{ *id, makeRelativePath(fs::canonical(path.parent_path() / fs::path(wide)), rootInputPath).u8string() });
		}

		fs::path lookupPath = path;
		lookupPath.replace_extension();
		
		dbColl.name = path.stem().u8string();
		dbColl.dir = makeRelativePath(lookupPath.parent_path(), rootInputPath).u8string() + "\\";

		lookupPath = makeRelativePath(lookupPath, rootInputPath);

		std::cout << lookupPath << std::endl;

		dbLookup.emplace(lookupPath.u8string(), std::move(dbColl));

		
	}

	std::cout << "Done scanning for databases." << std::endl;

	static const std::string kExtensions[] = {
		"adb",
		"mod",
		"odb",
		"sdb",
		"txd",
		"ssd",
		"lvl",
	};

	const std::string kRootInputDirStr(kRootInputDir.begin(), kRootInputDir.end());
	
	// Model will need absolute paths for textures.
	const std::string kRootOutputDirStr = fs::absolute(fs::path(kRootOutputDir.begin(), kRootOutputDir.end())).u8string();

	for (auto& kv : dbLookup)
	{
		const std::string_view name = kv.first;
		const drakan::DatabaseCollection& dbColl = kv.second;

		std::cout << "Loading " << name << std::endl;

		for (const std::string& ext : kExtensions)
		{
			const std::string relPath = std::string(name) + std::string(".") + ext;
			if (auto optDB = drakan::loadDatabase(kRootInputDirStr + relPath))
			{
				const std::string outputPath = kRootOutputDirStr + relPath + ".txt";
				const String entriesDumpStr = optDB->getEntriesDump();
				FileUnbufferedWriteStream(string_view(outputPath), EFileWriteOpenMode::RECREATE, true).write(asRawUChars(asArrayView(entriesDumpStr)));

				optDB->parse();

				const String dbDumpDir = string_view(kRootOutputDirStr) + string_view(kv.second.dir)
					+ string_view(kv.second.name) + " [root]\\"_sv;

				drakan::NodeDumpArg dumpArg{};
				dumpArg.dbMeta = &dbColl;
				dumpArg.db = &*optDB;
				dumpArg.outputRootDir = string_view(kRootOutputDirStr);
				dumpArg.dbRootDir = dbDumpDir;
				dumpArg.dir = string_view(dbDumpDir);
				optDB->rootDir.dumpNode(dumpArg);
			}
		}

	}

	return 0;
}
catch (const std::runtime_error& ex)
{
	fprintf(stderr, "Runtime Exception: %s\n", ex.what());
	return 1;
}